﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Gencode.VECore.History
{
    public class HistoryManager
    {
        ShapeList shapeList;

        List<Command> historyList;
        int nextUndo;

        public HistoryManager(ShapeList shapeList)
        {
            this.shapeList = shapeList;

            ClearHistory();
        }

        /// <summary>
        /// Return true if Undo operation is available
        /// </summary>
        public bool CanUndo
        {
            get
            {
                // If the NextUndo pointer is -1, no commands to undo
                if (nextUndo < 0 ||
                    nextUndo > historyList.Count - 1)   // precaution
                {
                    return false;
                }

                return true;
            }
        }

        /// <summary>
        /// Return true if Redo operation is available
        /// </summary>
        public bool CanRedo
        {
            get
            {
                // If the NextUndo pointer points to the last item, no commands to redo
                if (nextUndo == historyList.Count - 1)
                {
                    return false;
                }

                return true;
            }
        }

        /// <summary>
        /// Clear History
        /// </summary>
        public void ClearHistory()
        {
            historyList = new List<Command>();
            nextUndo = -1;
        }

        /// <summary>
        /// Add new command to history.
        /// Called by client after executing some action.
        /// </summary>
        /// <param name="command"></param>
        public void AddCommandToHistory(Command command)
        {
            // Purge history list
            this.TrimHistoryList();

            // Add command and increment undo counter
            historyList.Add(command);

            nextUndo++;
        }

        /// <summary>
        /// Undo
        /// </summary>
        public void Undo()
        {
            if ( ! CanUndo )
            {
                return;
            }

            // Get the Command object to be undone
            Command command = historyList[nextUndo];

            // Execute the Command object's undo method
            command.Undo(shapeList);

            // Move the pointer up one item
            nextUndo--;
        }

        /// <summary>
        /// Redo
        /// </summary>
        public void Redo()
        {
            if ( ! CanRedo )
            {
                return;
            }

            // Get the Command object to redo
            int itemToRedo = nextUndo + 1;
            Command command = historyList[itemToRedo];

            // Execute the Command object
            command.Redo(shapeList);

            // Move the undo pointer down one item
            nextUndo++;
        }

        private void TrimHistoryList()
        {
            // Exit if no items in History list
            if (historyList.Count == 0)
            {
                return;
            }

            // Exit if NextUndo points to last item on the list
            if (nextUndo == historyList.Count - 1)
            {
                return;
            }

            // Purge all items below the NextUndo pointer
            for (int i = historyList.Count - 1; i > nextUndo; i--)
            {
                historyList.RemoveAt(i);
            }
        }
    }
}
