﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media;
using System.Windows;

namespace Gencode.VECore.VectorShape
{
	public class VECEllipse : VECShape
	{
		System.Windows.Media.EllipseGeometry _definingGeometry;
        /// <summary>
        /// Serializer object description
        /// </summary>
        private const string serEllipse = "Elli";

		public VECEllipse()
			: this(0, 0, 1, 1)
		{
		}

		public VECEllipse(double x, double y, double radiusx, double radiusy) :
			base()
		{
			_definingGeometry = new System.Windows.Media.EllipseGeometry();
			_definingGeometry.Center = new System.Windows.Point(x, y);
			_definingGeometry.RadiusX = radiusx;
			_definingGeometry.RadiusY = radiusy;

			this.MouseLeftButtonUp += new System.Windows.Input.MouseButtonEventHandler(MouseLeftUp);

			this.Focusable = true;
		}

        /// <summary>
        /// Clone this instance
        /// </summary>
        public override VECShape Clone()
        {
            VECEllipse drawEllipse = new VECEllipse();
            drawEllipse.Rectangle = this.Rectangle;

            CopyShapeProperies(drawEllipse);
            return drawEllipse;
        }

		protected override System.Windows.Media.Geometry DefiningGeometry
		{
			get
			{
				if (_definingGeometry == null)
					_definingGeometry = new System.Windows.Media.EllipseGeometry();
				return _definingGeometry;
			}
		}

		protected override void OnRender(DrawingContext drawingContext)
		{
			base.OnRender(drawingContext);

			if (Selected)
			{
				Pen p = new Pen(Brushes.Blue, 1);
				drawingContext.DrawEllipse(Brushes.LightBlue, p, new Point(X, Y), 2.5, 2.5);
				drawingContext.DrawEllipse(Brushes.LightBlue, p, new Point(X + Width, Y), 2.5, 2.5);
				drawingContext.DrawEllipse(Brushes.LightBlue, p, new Point(X, Y + Height), 2.5, 2.5);
				drawingContext.DrawEllipse(Brushes.LightBlue, p, new Point(X + Width, Y + Height), 2.5, 2.5);
			}
		}



		public double X
		{
			get
			{
				return _definingGeometry.Bounds.X;
			}
		}

		public double Y
		{
			get
			{
				return _definingGeometry.Bounds.Y;
			}
		}

		public new double Height
		{
			get
			{
				return _definingGeometry.Bounds.Height;
			}
		}

		public new double Width
		{
			get
			{
				return _definingGeometry.Bounds.Width;
			}
		}

        internal Rect Rectangle
        {
            get
            {
                return _definingGeometry.Bounds;
            }
            set
            {
                _definingGeometry.Center = new Point(value.X + (value.Width / 2), value.Y + (value.Height / 2));
                _definingGeometry.RadiusX = value.Width / 2;
                _definingGeometry.RadiusY = value.Height / 2;
            }
        }


		private void MouseLeftUp(object sender, System.Windows.Input.MouseButtonEventArgs e)
		{
			if (CanvasDocument.SelectedToolType == Gencode.VECore.ToolType.SelectionTool)
			{
				if (!Selected)
				{
					Selected = true;
				}
			}
		}



		/// <summary>
		/// Get number of handles
		/// </summary>
		public override int HandleCount
		{
			get
			{
				return 4;
			}
		}


		/// <summary>
		/// Get handle point by 1-based number
		/// </summary>
		/// <param name="handleNumber"></param>
		/// <returns></returns>
		public override Point GetHandle(int handleNumber)
		{
			double x, y;
			Rect _rectangle = _definingGeometry.Bounds;
			x = _rectangle.X;
			y = _rectangle.Y;

			switch (handleNumber)
			{
				case 1:
					x = _rectangle.X;
					y = _rectangle.Y;
					break;
				case 2:
					x = _rectangle.Right;
					y = _rectangle.Y;
					break;
				case 3:
					x = _rectangle.Right;
					y = _rectangle.Bottom;
					break;
				case 4:
					x = _rectangle.X;
					y = _rectangle.Bottom;
					break;
			}

			return new Point(x, y);

		}

		/// <summary>
		/// Hit test.
		/// Return value: -1 - no hit
		///                0 - hit anywhere
		///                > 1 - handle number
		/// </summary>
		/// <param name="point"></param>
		/// <returns></returns>
		public override int HitTest(Point point)
		{
			if (Selected)
			{
				for (int i = 1; i <= HandleCount; i++)
				{
					if (GetHandleRectangle(i).Contains(point))
						return i;
				}
			}

			if (PointInObject(point))
				return 0;

			return -1;
		}

		protected override bool PointInObject(Point point)
		{
			Rect _rectangle = _definingGeometry.Bounds;
			return _rectangle.Contains(point);
		}

		/// <summary>
		/// Move handle to new point (resizing)
		/// </summary>
		/// <param name="point"></param>
		/// <param name="handleNumber"></param>
		public override void MoveHandle(Point point, int handleNumber)
		{
			Rect _rectangle = _definingGeometry.Bounds;
			double left = _rectangle.Left;
			double top = _rectangle.Top;
			double right = _rectangle.Right;
			double bottom = _rectangle.Bottom;

			switch (handleNumber)
			{
				case 1:
					left = point.X;
					top = point.Y;
					break;
				case 2:
					right = point.X;
					top = point.Y;
					break;
				case 3:
					right = point.X;
					bottom = point.Y;
					break;
				case 4:
					left = point.X;
					bottom = point.Y;
					break;
			}

			SetRectangle(left, top, right - left, bottom - top);

            CallChangedHandler();
		}

		protected void SetRectangle(double x, double y, double width, double height)
		{
			try
			{
				Rect _rectangle = _definingGeometry.Bounds;
				_rectangle.X = x;
				_rectangle.Y = y;
				_rectangle.Width = width;
				_rectangle.Height = height;

				_definingGeometry.Center = new Point(_rectangle.X + (_rectangle.Width / 2), _rectangle.Y + (_rectangle.Height / 2));
				_definingGeometry.RadiusX = _rectangle.Width / 2;
				_definingGeometry.RadiusY = _rectangle.Height / 2;
				this.InvalidateVisual();
			}
			catch { }
		}

        public override void Move(double deltaX, double deltaY)
        {
            _definingGeometry.Center = new Point(_definingGeometry.Center.X + deltaX, _definingGeometry.Center.Y + deltaY);

            this.InvalidateVisual();

            CallMovedHandler();
        }

		/// <summary>
		/// Get cursor for the handle
		/// </summary>
		/// <param name="handleNumber"></param>
		/// <returns></returns>
		public override System.Windows.Input.Cursor GetHandleCursor(int handleNumber)
		{
			switch (handleNumber)
			{
				case 1:
					return System.Windows.Input.Cursors.SizeNWSE;
				case 2:
					return System.Windows.Input.Cursors.SizeNESW;
				case 3:
					return System.Windows.Input.Cursors.SizeNWSE;
				case 4:
					return System.Windows.Input.Cursors.SizeNESW;
				default:
					return System.Windows.Input.Cursors.Arrow;
			}
		}

        public override bool IntersectsWith(Rect rectangle)
        {
            return Rectangle.IntersectsWith(rectangle);
        }

        public override void SaveToStream(System.Runtime.Serialization.SerializationInfo info, int orderNumber)
        {
            info.AddValue(
                String.Format(System.Globalization.CultureInfo.InvariantCulture,
                "{0}{1}",
                serEllipse, orderNumber),
                Rectangle);

            base.SaveToStream(info, orderNumber);
        }

        public override void LoadFromStream(System.Runtime.Serialization.SerializationInfo info, int orderNumber)
        {
            Rectangle = (Rect)info.GetValue(
                String.Format(System.Globalization.CultureInfo.InvariantCulture,
                "{0}{1}",
                serEllipse, orderNumber),
                typeof(Rect));

            base.LoadFromStream(info, orderNumber);
        }






        //public override void Draw(DrawingContext drawingContext)
        //{
        //    drawingContext.DrawEllipse(Fill, new Pen(Stroke, StrokeThickness), _definingGeometry.Center, _definingGeometry.RadiusX, _definingGeometry.RadiusY);
        //    base.Draw(drawingContext);
        //}

        //public override void DrawTracker(DrawingContext drawingContext)
        //{
        //    base.DrawTracker(drawingContext);
        //}
	}
}
