﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WickedFlame.Core.Shapes;
using System.Windows.Media;
using System.Collections;
using System.ComponentModel;

namespace WickedFlame.Core
{
    
    

    /// <summary>
    /// List of graphic objects
    /// </summary>
    [Serializable]
    public class ShapeList : IEnumerable, INotifyPropertyChanged
    {
        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected void NotifyPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        VisualCollection _shapeList;

        public ShapeList(Visual parent)
        {
            _shapeList = new VisualCollection(parent);
        }

        /// <summary>
        /// Clear all objects in the list
        /// </summary>
        /// <returns>
        /// true if at least one object is deleted
        /// </returns>
        public bool Clear()
        {
            bool result = (_shapeList.Count > 0);
            _shapeList.Clear();
            return result;
        }

        /// <summary>
        /// 
        /// </summary>
        public int Count
        {
            get
            {
                return _shapeList.Count;
            }
        }

        public int IndexOf(DrawableShape shape)
        {
            return _shapeList.IndexOf(shape);
        }

        public DrawableShape this[int index]
        {
            get
            {
                if (index < 0 || index >= _shapeList.Count)
                    return null;

                return _shapeList[index] as DrawableShape;
            }
        }


        /// <summary>
        /// SelectedCount and GetSelectedObject allow to read
        /// selected objects in the loop
        /// </summary>
        public int SelectionCount
        {
            get
            {
                int n = 0;

                foreach (VectorShape shape in Selection)
                {
                    n++;
                }

                return n;
            }
        }


        /// <summary>
        /// Returns INumerable object which may be used for enumeration
        /// of selected objects.
        /// 
        /// Note: returning IEnumerable<VECShape> breaks CLS-compliance
        /// (assembly CLSCompliant = true is removed from AssemblyInfo.cs).
        /// To make this program CLS-compliant, replace 
        /// IEnumerable<VECShape> with IEnumerable. This requires
        /// casting to object at runtime.
        /// </summary>
        /// <value></value>
        public IEnumerable<VectorShape> Selection
        {
            get
            {
                foreach (VectorShape shape in _shapeList)
                {
                    if (shape.IsSelected)
                    {
                        yield return shape;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the shape that is selected. Retruns null if more than 1 shape is selcted.
        /// </summary>
        public VectorShape SelectedShape
        {
            get
            {
                VectorShape shape = null;

                foreach (VectorShape s in _shapeList)
                {
                    if (s.IsSelected)
                    {
                        if (shape == null)
                            shape = s;
                        else
                            return null;
                    }
                }

                return shape;
            }
        }

        public VectorShape TopShape
        {
            get
            {
                if (_shapeList.Count <= 0)
                    return null;

                return _shapeList[_shapeList.Count - 1] as VectorShape;
            }
        }

        /// <summary>
        /// Inserts a VECShape on the top of the list
        /// </summary>
        /// <param name="shape"></param>
        public void Add(DrawableShape shape)
        {
            _shapeList.Add(shape);
        }

        /// <summary>
        /// Insert object to specified place.
        /// Used for Undo.
        /// </summary>
        internal void Insert(int index, VectorShape shape)
        {
            if (index >= 0 || index < _shapeList.Count)
            {
                _shapeList.Insert(index, shape);
            }
        }

        /// <summary>
        /// Removes a VECShape from the list
        /// </summary>
        /// <param name="shape"></param>
        internal void Remove(VectorShape shape)
        {
            if (_shapeList.Contains(shape))
                _shapeList.Remove(shape);
        }

        /// <summary>
        /// Removes all selected objects
        /// </summary>
        internal void RemoveSelection()
        {
            foreach (VectorShape shape in _shapeList)
            {
                if (shape.IsSelected)
                    _shapeList.Remove(shape);
            }
            SelectionChanged();
        }

        /// <summary>
        /// Replace shape in specified shape.
        /// Used for Undo.
        /// </summary>
        internal void Replace(int index, VectorShape shape)
        {
            if (index >= 0 && index < _shapeList.Count)
            {
                _shapeList.RemoveAt(index);
                _shapeList.Insert(index, shape);
            }
        }

        /// <summary>
        /// Replace shape specified shape.
        /// Retruns false if failed to repalce
        /// DO NOT CALL THIS MESSAGE FROM OUTSIDE OF THE LAYER! CALL layer.Replace(...) INSTEAD!
        /// </summary>
        internal bool Replace(VectorShape originalShape, VectorShape newShape)
        {
            if (_shapeList.Contains(originalShape))
            {
                int index = _shapeList.IndexOf(originalShape);
                _shapeList[index] = null;
                _shapeList[index] = newShape;
                return true;
            }

            return false;
        }

        /// <summary>
        /// Remove object by index.
        /// Used for Undo.
        /// </summary>
        public void RemoveAt(int index)
        {
            _shapeList.RemoveAt(index);
        }

        /// <summary>
        /// Delete last added object from the list
        /// (used for Undo operation).
        /// </summary>
        public void DeleteLastAddedObject()
        {
            if ( _shapeList.Count > 0 )
            {
                //_shapeList.RemoveAt(0);
                _shapeList.RemoveAt(_shapeList.Count - 1);
            }
        }

        /// <summary>
        /// Selects all shapes that intersect with the rectangle
        /// </summary>
        /// <param name="rectangle"></param>
        public void SelectShapes(System.Windows.Rect rectangle)
        {
            UnselectAll();

            foreach (VectorShape shape in _shapeList)
            {
                if (shape != null)
                    if (shape.IntersectsWith(rectangle))
                        shape.IsSelected = true;
            }

            SelectionChanged();
        }

        public void UnselectAll()
        {
            foreach (DrawableShape shape in _shapeList)
            {
                shape.IsSelected = false;
            }

            DrawingCanvas.DrawTrackers();
            SelectionChanged();
        }

        public void SelectAll()
        {
            foreach (DrawableShape shape in _shapeList)
            {
                shape.IsSelected = true;
            }

            DrawingCanvas.DrawTrackers();
            SelectionChanged();
        }

        public void SelectionChanged()
        {
            NotifyPropertyChanged("Selection");
            NotifyPropertyChanged("SelectedShape");
        }

        /// <summary>
        /// Move selected items to front (beginning of the list)
        /// </summary>
        /// <returns>
        /// true if at least one object is moved
        /// </returns>
        public bool MoveSelectionToFront()
        {
            int n;
            int i;
            var tempList = new List<DrawableShape>();
            n = _shapeList.Count;

            // Read source list in reverse order, add every selected item
            // to temporary list and remove it from source list
            for (i = n - 1; i >= 0; i--)
            {
                if ((_shapeList[i] as DrawableShape).IsSelected)
                {
                    tempList.Add(_shapeList[i] as DrawableShape);
                    _shapeList.RemoveAt(i);
                }
            }

            // Read temporary list in direct order and insert every item
            // to the beginning of the source list
            n = tempList.Count;

            for (i = 0; i < n; i++)
            {
                _shapeList.Insert(0, tempList[i]);
            }

            SelectionChanged();

            return (n > 0);
        }

        /// <summary>
        /// Move selected items to back (end of the list)
        /// </summary>
        /// <returns>
        /// true if at least one object is moved
        /// </returns>
        public bool MoveSelectionToBack()
        {
            int n;
            int i;
            List<DrawableShape> tempList;

            tempList = new List<DrawableShape>();
            n = _shapeList.Count;

            // Read source list in reverse order, add every selected item
            // to temporary list and remove it from source list
            for (i = n - 1; i >= 0; i--)
            {
                if ((_shapeList[i] as DrawableShape).IsSelected)
                {
                    tempList.Add(_shapeList[i] as DrawableShape);
                    _shapeList.RemoveAt(i);
                }
            }

            // Read temporary list in reverse order and add every item
            // to the end of the source list
            n = tempList.Count;

            for (i = n - 1; i >= 0; i--)
            {
                _shapeList.Add(tempList[i]);
            }

            SelectionChanged();

            return (n > 0);
        }

        #region IEnumerable Members

        public IEnumerator GetEnumerator()
        {
            return new ShapeListEnumerator(this);
        }

        #endregion
    }

    public class ShapeListEnumerator : IEnumerator
    {
        ShapeList _list;
		int index;

        public ShapeListEnumerator(ShapeList list)
		{
			_list = list;
            index = -1;
		}

		#region IEnumerator Members

        public object Current
        {
            get
            {
                return _list[index];
            }
        }

        public bool MoveNext()
        {
            index++;

            return (index < _list.Count);
        }

        public void Reset()
        {
            index = -1;
        }
        //public bool MoveNext()
        //{
        //    if (index > 0)
        //    {
        //        index--;
        //        return true;
        //    }

        //    return false;
        //}

        //public void Reset()
        //{
        //    index = _list.Count;
        //}

		#endregion
    }
}
